/* Copyright (c) 2011, Stanislaw Adaszewski
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Stanislaw Adaszewski nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL STANISLAW ADASZEWSKI BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include <stdio.h>

#define __USE_GNU

#include <dlfcn.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#define _FCNTL_H
#include <bits/fcntl.h>

static int (*open64_orig)(const char*, int, ...);
static FILE* (*fopen64_orig)(const char*, const char*);
static int (*access_orig)(const char*, int);
static void *orig_stat;
static void *orig_lstat;
static void *orig_stat64;
static void *orig_lstat64;
static void *orig___xstat;
static void *orig___lxstat;
static void *orig___xstat64;
static void *orig___lxstat64;
static void *orig_catopen;

static char *union_spec[16] = {"", 0};

static char loaded;

static char debug;

static void info(const char *fmt, ...) {
	if (!debug)
		return;

	va_list ap;
	va_start (ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

void __attribute__((constructor)) init() {
	info("init()\n");
	if (loaded)
		return;

	debug = (getenv("UNIONFS_DEBUG") != 0);

	open64_orig = (int(*)(const char*, int, ...)) dlsym(RTLD_NEXT, "open64");
	fopen64_orig = (FILE*(*)(const char*, const char*)) dlsym(RTLD_NEXT, "fopen64");
	access_orig = (int(*)(const char*, int)) dlsym(RTLD_NEXT, "access");
	orig_stat = dlsym(RTLD_NEXT, "stat");
	orig_lstat = dlsym(RTLD_NEXT, "lstat");
	orig_stat64 = dlsym(RTLD_NEXT, "stat64");
	orig_lstat64 = dlsym(RTLD_NEXT, "lstat64");
	orig___xstat = dlsym(RTLD_NEXT, "__xstat");
	orig___lxstat = dlsym(RTLD_NEXT, "__lxstat");
	orig___xstat64 = dlsym(RTLD_NEXT, "__xstat64");
	orig___lxstat64 = dlsym(RTLD_NEXT, "__lxstat64");
	orig_catopen = dlsym(RTLD_NEXT, "catopen");

	info("open64_orig=%llu\n", open64_orig);
	info("fopen64_orig=%llu\n", fopen64_orig);
	
	char *s = strdup(getenv("UNIONFS"));
	if (!s)
		return;
	char **us = &union_spec[2];
	while (*s) {
		if (!*us)
			*us = s;
		if (*s == ':') {
			*s = 0;
			us++;
		}
		s++;
	}

	loaded = 1;
}

void __attribute__((destructor)) finish() {
	if (!loaded)
		return;

	free(union_spec[2]);
}

static char* find_file(const char *pathname) {
	init();

	info("find_file(): %s\n", pathname);

	char *buf = (char*) malloc(1024);
	
	char **us;
	
	if (pathname[0] == '/')
		us = &union_spec[2];
	else
		us = &union_spec[0];
		
	while (*us) {
		int n;
		if ((n = snprintf(buf, 1024, "%s%s", *us, pathname)) > 1024) {
			buf = (char*) realloc(buf, n);
			snprintf(buf, n, "%s%s", *us, pathname);
		}
		
		info("buf=%s\n", buf);
		
		int fd;
		fd = open64_orig(buf, O_RDONLY);
		if (fd != -1) {
			close(fd);
			return buf;
		}
		
		us++;
	}
	
	free(buf);
	return 0;
}

int open64(const char *pathname, int flags, ...) {
	info("open64(): %s, %d\n", pathname, flags);
	
	char *buf = find_file(pathname);
	if (!buf)
		buf = (char*) pathname;
		
	int ret;
	
	if (flags & O_CREAT) {	
		va_list ap;
		va_start(ap, flags);
		__mode_t mode = va_arg(ap, mode_t);
		ret = open64_orig(buf, flags, mode);
		va_end(ap);
	} else {
		ret = open64_orig(buf, flags);
	}
	
	if (buf != pathname)
		free(buf);
	
	return ret;
}

FILE* fopen64(const char *pathname, const char *mode) {
	info("fopen64(): %s, %s\n", pathname, mode);
	
	char *buf = find_file(pathname);
	if (!buf)
		buf = (char*) pathname;
	
	FILE *ret = fopen64_orig(buf, mode);
	
	if (buf != pathname)
		free(buf);
	
	return ret;
}

int __attribute__((weak, alias("open64"))) open(const char*, int, ...);
FILE* __attribute__((weak, alias("fopen64"))) fopen(const char*, const char*);

int access(const char *pathname, int mode) {
	info("access(): %s, %d\n", pathname, mode);

	char *buf = find_file(pathname);
	if (!buf)
		buf = (char*) pathname;
		
	int ret = access_orig(buf, mode);
		
	if (buf != pathname)
		free(buf);
		
	return ret;
}

#define WRAP2(ret_type, name, arg0_type, arg1_type, pathname_arg) \
	ret_type name(arg0_type arg0, arg1_type arg1) { \
		info("%s(): %s\n", #name, pathname_arg); \
		char *buf = find_file(pathname_arg); \
		if (buf) \
			pathname_arg = buf; \
		ret_type ret = ((ret_type(*)(arg0_type, arg1_type)) orig_##name)(arg0, arg1); \
		free(buf); \
		return ret; \
	}
	
#define WRAP3(ret_type, name, arg0_type, arg1_type, arg2_type, pathname_arg) \
	ret_type name(arg0_type arg0, arg1_type arg1, arg2_type arg2) { \
		info("%s(): %s\n", #name, pathname_arg); \
		char *buf = find_file(pathname_arg); \
		if (buf) \
			pathname_arg = buf; \
		ret_type ret = ((ret_type(*)(arg0_type, arg1_type, arg2_type)) orig_##name)(arg0, arg1, arg2); \
		free(buf); \
		return ret; \
	}
	
#define WRAP4(ret_type, name, arg0_type, arg1_type, arg2_type, arg3_type, pathname_arg) \
	ret_type name(arg0_type arg0, arg1_type arg1, arg2_type arg2, arg3_type arg3) { \
		info("%s(): %s\n", #name, pathname_arg); \
		char *buf = find_file(pathname_arg); \
		if (buf) \
			pathname_arg = buf; \
		ret_type ret = ((ret_type(*)(arg0_type, arg1_type, arg2_type, arg3_type)) orig_##name)(arg0, arg1, arg2, arg3); \
		free(buf); \
		return ret; \
	}
	
struct stat;
struct stat64;

WRAP2(int, stat, const char*, struct stat*, arg0)
WRAP2(int, lstat, const char*, struct stat*, arg0)

WRAP2(int, stat64, const char*, struct stat64*, arg0)
WRAP2(int, lstat64, const char*, struct stat64*, arg0)

WRAP3(int, __xstat, int, const char *, struct stat*, arg1)
WRAP3(int, __lxstat, int, const char*, struct stat*, arg1)

WRAP3(int, __xstat64, int, const char *, struct stat64*, arg1)
WRAP3(int, __lxstat64, int, const char*, struct stat64*, arg1)

#include <nl_types.h>

WRAP2(nl_catd, catopen, const char*, int, arg0);
